@@ -1,7 +1,10 @@
 import json
+import logging
+from collections import Counter
 from functools import partial
 from pathlib import Path
-from typing import Any, Dict, Optional
+from typing import Any, AsyncIterable, Dict, List, Optional, Tuple
+from uuid import UUID
 
 from kilroy_module_pytorch_py_sdk import (
     BasicModule,
@@ -27,6 +30,8 @@ from kilroy_module_huggingface.models import HuggingfaceLanguageModel
 from kilroy_module_huggingface.modules.base import HuggingfaceModule
 from kilroy_module_huggingface.tokenizer import HuggingfaceTokenizer
 
+logger = logging.getLogger(__name__)
+
 
 class Params(SerializableModel):
     model_name: str
@@ -108,7 +113,7 @@ class BasicHuggingfaceModule(BasicModule, HuggingfaceModule[State]):
     @staticmethod
     async def _build_reports() -> ReportsState:
         return ReportsState(
-            epoch_reinforced_scores=[], epoch_supervised_losses=[]
+            step_reinforced_scores=[], step_supervised_losses=[]
         )
 
     async def _build_default_state(self) -> State:
@@ -127,7 +132,7 @@ class BasicHuggingfaceModule(BasicModule, HuggingfaceModule[State]):
             codec=await self._build_codec(params),
             results_cache={},
             batch_size=params.batch_size,
-            epoch=0,
+            step=0,
             metrics=await self._build_metrics(),
             reports=await self._build_reports(),
         )
@@ -170,9 +175,9 @@ class BasicHuggingfaceModule(BasicModule, HuggingfaceModule[State]):
             else None,
             "schedulers_params": state.schedulers_params,
             "batch_size": state.batch_size,
-            "epoch": state.epoch,
-            "epoch_supervised_losses": state.reports.epoch_supervised_losses,
-            "epoch_reinforced_scores": state.reports.epoch_reinforced_scores,
+            "step": state.step,
+            "step_supervised_losses": state.reports.step_supervised_losses,
+            "step_reinforced_scores": state.reports.step_reinforced_scores,
         }
 
     @staticmethod
@@ -223,6 +228,7 @@ class BasicHuggingfaceModule(BasicModule, HuggingfaceModule[State]):
             category=state_dict["optimizer_type"],
             default=partial(cls._build_optimizer, params, model),
             parameters=model.parameters(),
+            **params.optimizers_params.get(params.optimizer_type, {}),
         )
 
     @classmethod
@@ -243,7 +249,7 @@ class BasicHuggingfaceModule(BasicModule, HuggingfaceModule[State]):
         params: Params,
         optimizer: Optimizer,
     ) -> Optional[Scheduler]:
-        if state_dict["scheduler_type"] is None:
+        if state_dict.get("scheduler_type") is None:
             return None
         return await cls._load_generic(
             directory / "scheduler",
@@ -251,6 +257,7 @@ class BasicHuggingfaceModule(BasicModule, HuggingfaceModule[State]):
             category=state_dict["scheduler_type"],
             default=partial(cls._build_scheduler, params, optimizer),
             optimizer=await optimizer.get(),
+            **params.schedulers_params.get(params.scheduler_type, {}),
         )
 
     @classmethod
@@ -261,6 +268,7 @@ class BasicHuggingfaceModule(BasicModule, HuggingfaceModule[State]):
             directory / "generator",
             Generator,
             default=partial(cls._build_generator, params),
+            **params.generator_params,
         )
 
     @classmethod
@@ -269,6 +277,7 @@ class BasicHuggingfaceModule(BasicModule, HuggingfaceModule[State]):
             directory / "codec",
             Codec,
             default=partial(cls._build_codec, params),
+            **params.codec_params,
         )
 
     async def _load_saved_state(self, directory: Path) -> State:
@@ -292,10 +301,49 @@ class BasicHuggingfaceModule(BasicModule, HuggingfaceModule[State]):
             codec=await self._load_codec(directory, params),
             results_cache={},
             batch_size=state_dict["batch_size"],
-            epoch=state_dict["epoch"],
+            step=state_dict["step"],
             metrics=await self._build_metrics(),
             reports=ReportsState(
-                epoch_supervised_losses=state_dict["epoch_supervised_losses"],
-                epoch_reinforced_scores=state_dict["epoch_reinforced_scores"],
+                step_supervised_losses=state_dict["step_supervised_losses"],
+                step_reinforced_scores=state_dict["step_reinforced_scores"],
             ),
         )
+
+    async def generate(
+        self, n: int, dry: bool
+    ) -> AsyncIterable[Tuple[UUID, Dict[str, Any]]]:
+        logger.info(f"Generating {n} posts...")
+
+        async for post_id, post in super().generate(n, dry):
+            logger.info(f"Generated post {str(post_id)}.")
+            yield post_id, post
+
+        logger.info("Finished generating posts.")
+
+    async def fit_posts(
+        self, posts: AsyncIterable[Tuple[Dict[str, Any], float]]
+    ) -> None:
+        logger.info("Fitting posts...")
+
+        posts_counter = Counter(posts=0)
+
+        async def count_posts(
+            _posts: AsyncIterable[Tuple[Dict[str, Any], float]],
+            _posts_counter: Counter,
+        ) -> AsyncIterable[Tuple[Dict[str, Any], float]]:
+            async for post, score in _posts:
+                _posts_counter["posts"] += 1
+                yield post, score
+
+        await super().fit_posts(count_posts(posts, posts_counter))
+        logger.info(f"Finished fitting {posts_counter['posts']} posts.")
+
+    async def fit_scores(self, scores: List[Tuple[UUID, float]]) -> None:
+        logger.info(f"Fitting {len(scores)} scores...")
+        await super().fit_scores(scores)
+        logger.info("Finished fitting scores.")
+
+    async def step(self) -> None:
+        logger.info("Performing step...")
+        await super().step()
+        logger.info("Finished performing step.")
